/*******************************************************************************
 Copyright(c) 2017 Jasem Mutlaq. All rights reserved.

 Simple SkyCommander DSC Driver

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License version 2 as published by the Free Software Foundation.
 .
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.
 .
 You should have received a copy of the GNU Library General Public License
 along with this library; see the file COPYING.LIB.  If not, write to
 the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 Boston, MA 02110-1301, USA.
*******************************************************************************/

#pragma once

#include "inditelescope.h"

class SkyCommander : public INDI::Telescope
{
  public:
    SkyCommander();
    virtual ~SkyCommander() = default;

  protected:
    virtual const char *getDefaultName();
    virtual bool Handshake();
    virtual bool ReadScopeStatus();
};
