/*!
 * \file DriverCommon.cpp
 *
 * \author Roger James
 * \date 28th January 2014
 *
 */

#include "indilogger.h"

namespace INDI
{
namespace AlignmentSubsystem
{
int DBG_ALIGNMENT = INDI::Logger::getInstance().addDebugLevel("Alignment Subsystem", "ALIGNMENT");

} // namespace AlignmentSubsystem
} // namespace INDI
